{-|
Module      : Test.API.Notifications
Description : Tests for the notification mechanism (central exchange, dispatcher, websockets)
Copyright   : (c) CNRS, 2017-Present
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

-}

{-# OPTIONS_GHC -Wno-orphans  #-}

{-# LANGUAGE ScopedTypeVariables #-}

    
module Test.API.Notifications (
    tests
  ) where

import Control.Concurrent (threadDelay)
import Control.Concurrent.Async (withAsync)
import Control.Concurrent.STM.TChan
import Control.Lens ((^.))
import Control.Monad.STM (atomically)
import Data.Aeson qualified as Aeson
import Data.Maybe (isJust)
import Gargantext.Core.Config (gc_notifications_config)
import Gargantext.Core.Notifications.CentralExchange qualified as CE
import Gargantext.Core.Notifications.CentralExchange.Types qualified as CET
import Gargantext.Core.Notifications.Dispatcher.Types qualified as DT
import Gargantext.System.Logging (logMsg, LogLevel(DEBUG), withLogger)
import Network.WebSockets qualified as WS
import Prelude
import System.Timeout qualified as Timeout
import Test.API.Setup (SpecContext(..), withTestDBAndPort)
import Test.Database.Types (test_config)
import Test.Hspec
import Test.Instances ()
import Test.Utils.Notifications



-- tests :: D.Dispatcher -> Spec
tests :: Spec
tests = sequential $ aroundAll withTestDBAndPort $ do
  describe "Notifications" $ do
    it "simple WS notification works" $ \(SpecContext testEnv port _app _) -> do
      let nc = (test_config testEnv) ^. gc_notifications_config
      
      let topic = DT.UpdateTree 0
      tchan <- newTChanIO :: IO (TChan (Maybe DT.Notification))
      -- setup a websocket connection
      let wsConnect =
            withWSConnection ("127.0.0.1", port) $ \conn -> do
              -- We wait a bit before the server settles
              -- threadDelay (100 * millisecond)
              withLogger () $ \ioL ->
                logMsg ioL DEBUG $ "[wsConnect] subscribing topic: " <> show topic
              WS.sendTextData conn $ Aeson.encode (DT.WSSubscribe topic)
              d <- WS.receiveData conn
              let dec = Aeson.decode d :: Maybe DT.Notification
              atomically $ writeTChan tchan dec
            
      -- wait a bit to settle
      threadDelay (100 * millisecond)

      withAsync wsConnect $ \_a -> do
        -- wait a bit to connect
        threadDelay (500 * millisecond)

        let nodeId = 0
        CE.notify nc $ CET.UpdateTreeFirstLevel nodeId

        mTimeout <- Timeout.timeout (5 * 1000000) $ do
          md <- atomically $ readTChan tchan
  
          md `shouldBe` Just (DT.NUpdateTree nodeId)

        mTimeout `shouldSatisfy` isJust
